import os
import subprocess
from shutil import which
from abc import abstractmethod
from typing import Mapping, Type


class BackendBase:
    def __init__(self, name, config):
        super().__init__()
        self.name = name
        self.config = config
        self.env = self.config.get('env', {})
        self._command = self.config.get('command', name)
        if self._command and not which(self._command):
            raise BackendNotFoundError(
                f'{self._command!r}: command not found')
        slots = self.config.get('slots')
        if slots == 'auto':
            try:
                nv = subprocess.check_output(['nvidia-smi', '-L'])
                slots = len(nv.splitlines())
            except (FileNotFoundError, subprocess.CalledProcessError):
                slots = 1
        self.config['slots'] = slots

    def exec(self, *args, commit=True, shell=False, check=True):
        cmd = [self._command] + [str(a) for a in args]
        if not commit:
            print(' '.join(cmd))
            return None
        env = dict(os.environ, **self.env)
        if shell:
            subprocess.run(' '.join(cmd), shell=True, env=env, check=check)
            return None
        p = subprocess.run(
            cmd, capture_output=True, env=env, check=check)
        out = p.stdout.decode('utf-8').strip()
        out += p.stderr.decode('utf-8').strip()
        return out

    def backend_getset(self, key, value=None):
        pass

    def backend_command(self, command, commit=True):
        return self.exec(*command, commit=commit, check=False)

    @abstractmethod
    def backend_info(self):
        raise NotImplementedError

    @abstractmethod
    def backend_kill(self, args):
        raise NotImplementedError

    @abstractmethod
    def job_info(self, ids=None, filters=None):
        raise NotImplementedError

    @abstractmethod
    def full_info(
        self, ids=None, filters=None, extra_func=None, tqdm_disable=False
    ):
        raise NotImplementedError

    @abstractmethod
    def output(self, info, tail, shell=False):
        raise NotImplementedError

    @abstractmethod
    def add(self, command, gpus, slots, commit=True):
        raise NotImplementedError

    @abstractmethod
    def kill(self, info, commit=True):
        raise NotImplementedError

    @abstractmethod
    def remove(self, info, commit=True):
        raise NotImplementedError


BACKENDS: Mapping[str, Type[BackendBase]] = {}


def register_backend(name):
    def decorator(cls):
        cls.name = name
        BACKENDS[name] = cls
        return cls
    return decorator


class BackendNotFoundError(FileNotFoundError):
    pass



from abc import abstractmethod
from typing import Mapping, Type

from ..backends import BACKENDS, BackendNotFoundError


class ActionBase:
    def __init__(self, name, parser_kwargs):
        super().__init__()
        self.options = {}
        self.name = name
        self.parser_kwargs = parser_kwargs

    def transform_args(self, args):
        return args

    @abstractmethod
    def main(self, args):
        raise NotImplementedError

    def __call__(self, args, config):
        args = self.transform_args(args)
        backend = config['backend']
        try:
            backend_cls = BACKENDS[backend]
        except KeyError:
            print(f'Invalid backend: {backend}')
            backend_cls = BACKENDS['dummy']
        try:
            self.backend = backend_cls(backend, config)
        except BackendNotFoundError as e:
            print(f'Backend {backend!r} not available, reason: {e}.')
            return 1
        return self.main(args)


class DryActionBase(ActionBase):
    dry_options = {
        ('-d', '--dry-run'): {
            'action': 'store_true',
            'help': (
                'Do not actually perform the action, '
                'just print what would be done.'),
        },
    }

    def __init__(self, name, parser_kwargs):
        super().__init__(name, parser_kwargs)
        self.options.update(self.dry_options)

    def transform_args(self, args):
        args = super().transform_args(args)
        args.commit = not args.dry_run
        return args


_actions: Mapping[str, Type[ActionBase]] = {}
_aliases: Mapping[str, str] = {}
INFO = {
    'default': None,
    'actions': _actions,
    'aliases': _aliases,
}


def register_action(name, help=None, aliases=(), default=False):
    def decorator(cls):
        if name in INFO['actions']:
            raise ValueError(f'Action {name!r} already registered.')
        kwargs = {
            'name': name,
            'help': help,
            'aliases': aliases,
        }
        INFO['actions'][name] = cls(name, kwargs)
        INFO['aliases'].update({a: name for a in list(aliases) + [name]})
        if default:
            if INFO['default'] is not None:
                raise ValueError(
                    f'Default action already set to {INFO["default"]!r}.')
            INFO['default'] = name
        return cls
    return decorator



import os
import sys
from dataclasses import dataclass

from tqdm import tqdm as tqdm_


def tqdm(*args, disable=None, **kwargs):
    return tqdm_(*args, **kwargs, delay=1, disable=disable)


STATUSES = ['running', 'queued', 'success', 'failed', 'killed']


@dataclass
class FilterArgs:
    force_all: bool = False
    running: bool = False
    queued: bool = False
    success: bool = False
    failed: bool = False
    killed: bool = False

    @property
    def all(self):
        flags = all(not getattr(self, a) for a in STATUSES)
        return self.force_all or flags


STDIN_TTY = os.isatty(sys.stdin.fileno())
STDOUT_TTY = os.isatty(sys.stdout.fileno())
STDERR_TTY = os.isatty(sys.stderr.fileno())


def tail_lines(text, tail):
    if tail <= 0:
        return text
    return '\n'.join(text.split('\n')[-tail:])


def file_tail_lines(file, tail):
    if isinstance(file, str):
        try:
            with open(file, 'r', encoding='utf-8') as f:
                return tail_lines(f.read(), tail)
        except FileNotFoundError:
            return file
    return tail_lines(file.read(), tail)


def dict_merge(d, u):
    for k, v in u.items():
        if isinstance(v, dict):
            d[k] = dict_merge(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def dict_simplify(d, not_value=False):
    if not isinstance(d, dict):
        return d
    for k, v in list(d.items()):
        v = dict_simplify(v, not_value)
        if v == {} or (not_value and not v):
            del d[k]
        else:
            d[k] = v
    return d



import os
import json
from typing import Optional, Mapping, Any


try:
    SCREENDIR = os.environ['SCREENDIR']
except KeyError as e:
    raise ValueError('SCREENDIR not set.') from e


class Wrapper:
    def __init__(
        self, task_id=None, slots=1, gpus=0,
        command=None, additional_info: Optional[Mapping[str, Any]] = None,
    ):
        super().__init__()
        self.info = {
            **(additional_info or {}),
            'task_id': task_id or self._new_task_id(),
            'slots': slots,
            'gpus': gpus,
            'command': command,
            'status': 'queued',
        }

    def _new_task_id(self):
        files = os.listdir(SCREENDIR)
        jobs = [int(f.split('.')[0]) for f in files if f.startswith('tq.')]
        if not jobs:
            return 0
        return max(jobs) + 1

    def _update_info_file(self, info):
        path = f'{SCREENDIR}/tq.{self.info["task_id"]}'
        try:
            with open(path, 'r', encoding='utf-8') as f:
                old_info = json.load(f)
        except FileNotFoundError:
            old_info = {}
        with open(path, 'w', encoding='utf-8') as f:
            json.dump(
                {**old_info, **info}, f,
                indent=4, sort_keys=False, ensure_ascii=False)



def timedelta_format(delta, fmt, num_components):
    total_seconds = delta.total_seconds()
    components = {
        'w': 604800,
        'd': 86400,
        'h': 3600,
        'm': 60,
        's': 1,
    }
    if not all(c in components for c in fmt):
        raise ValueError(f'Invalid format: {fmt}')
    text = []
    leading_zeros = True
    for k, v in components.items():
        if k not in fmt:
            continue
        count = int(total_seconds // v)
        total_seconds -= count * v
        if leading_zeros and not count:
            continue
        leading_zeros = False
        if num_components is not None:
            if len(text) >= num_components:
                continue
        text.append(f'{count}{k}')
    return ''.join(text) or '0s'



import re
import sys
import textwrap
import argparse
import itertools

from ..common import tqdm, STDIN_TTY, FilterArgs
from .base import register_action, DryActionBase


@register_action('add', 'add jobs', aliases=['a'])
class AddAction(DryActionBase):
    add_options = {
        ('-G', '--gpus'): {
            'type': int,
            'default': None,
            'help': 'Number of GPUs required.',
        },
        ('-N', '--slots'): {
            'type': int,
            'default': None,
            'help': 'Number of slots required.',
        },
        ('-u', '--unique'): {
            'action': 'store_true',
            'help':
                'Only add unique commands. '
                'If a command is already in the queue, '
                'it will also be skipped.',
        },
        ('-i', '--interact'): {
            'action': 'store_true',
            'help': 'Interact with the added command that is running.',
        },
        ('-f', '--from-file'): {
            'type': str,
            'default': None,
            'help':
                'Read commands from a file, one per line. '
                'If "-", read from standard input.',
        },
        ('-s', '--separator'): {
            'type': str,
            'default': ',',
            'help': 'Separator for arguments from standard input.',
        },
        ('command', ): {
            'type': str,
            'nargs': argparse.REMAINDER,
            'help': 'The command to run.',
        },
    }

    def __init__(self, name, parser_kwargs):
        super().__init__(name, parser_kwargs)
        self.options.update(self.add_options)

    @staticmethod
    def _regex_extrapolate(texts, regex, extrapolator):
        new_texts = []
        for text in texts:
            scope_regex = re.compile(regex)
            scopes = scope_regex.findall(text)
            if not scopes:
                new_texts.append(text)
                continue
            values = itertools.product(*[extrapolator(s) for s in scopes])
            for v in values:
                replacer = lambda m, i=iter(v): str(next(i))
                h = scope_regex.sub(replacer, text)
                new_texts.append(h)
        return new_texts

    @classmethod
    def _extrapolate_ranges(cls, commands):
        def extrapolator(s):
            values = []
            for r in s[0].split(','):
                if not r:
                    continue
                if '-' in r:
                    start, end = r.split('-')
                    values += list(range(int(start), int(end) + 1))
                else:
                    values.append(int(r))
            return values
        regex = r'\[((?:\d+(?:-\d+)?)(?:,(?:\d+(-\d+)?))*)\]'
        return cls._regex_extrapolate(commands, regex, extrapolator)

    @classmethod
    def _extrapolate_sets(cls, commands):
        return cls._regex_extrapolate(
            commands, r'\{([^}]+)\}', lambda s: s.split(','))

    @staticmethod
    def _extrapolate_inputs(command, from_file, sep=','):
        if not STDIN_TTY:
            inputs = sys.stdin.read().split('\n')
        else:
            inputs = []
        if from_file == '-':
            commands = inputs
        elif from_file:
            with open(from_file, 'r', encoding='utf-8') as f:
                commands = f.read().split('\n')
        else:
            commands = []
        if command:
            commands += [' '.join(command)]
        if from_file == '-':
            # if we read from stdin for commands,
            # we can't read from stdin again for arguments
            return commands
        if not inputs:
            # nothing in stdin, so we can't extrapolate arguments
            return commands
        new_commands = []
        for line in inputs:
            line = line.strip()
            if not line:
                continue
            args = line.split(sep)
            for c in commands:
                # order reversed to avoid replacing "@1" in "@10"
                for j, a in reversed(list(enumerate(args))):
                    c = c.replace(f'@{j + 1}', a.strip())
                new_commands.append(c)
        return new_commands

    def main(self, args):
        commands = self._extrapolate_inputs(
            args.command, args.from_file, args.separator)
        commands = self._extrapolate_ranges(commands)
        commands = self._extrapolate_sets(commands)
        commands = [c.strip() for c in commands if c.strip()]
        if args.unique:
            info = self.backend.full_info(None, FilterArgs())
            queued_commands = [i['command'] for i in info]
            commands = list(dict.fromkeys(commands))
            skipped = [c for c in commands if c in queued_commands]
            commands = [c for c in commands if c not in queued_commands]
            if skipped:
                print('Skipped:')
                print(textwrap.indent('\n'.join(skipped), '  '))
        if not commands:
            print('No command to add.')
            if STDIN_TTY:
                print('Use "-f -" to read commands from stdin.')
            return
        ids = []
        for c in tqdm(commands, desc='add'):
            output = self.backend.add(
                c, args.gpus, args.slots, commit=args.commit)
            ids.append(output)
        if any(ids):
            print('Added:', ', '.join(ids))



import sys
import time
import platform
import textwrap
import functools
from datetime import datetime
from abc import abstractmethod

import tabulate
from blessed import Terminal

from ..common import tqdm, FilterArgs
from ..utils import timedelta_format
from .base import register_action
from .filter import FilterActionBase


term = Terminal()
COLOR_STATUS = {
    'running': term.green,
    'queued': term.yellow,
    'failed': term.red,
    'killed': term.orange,
    'success': term.blue,
}


class ReadActionBase(FilterActionBase):
    read_options = {
        ('-n', '--interval'): {
            'type': int,
            'default': None,
            'help': 'Refresh interval in seconds.',
        },
    }

    def __init__(self, name, parser_kwargs):
        super().__init__(name, parser_kwargs)
        self.options.update(self.read_options)

    @abstractmethod
    def format(self, args, tqdm_disable=False):
        raise NotImplementedError

    def loop(self, args):
        with term.hidden_cursor(), term.fullscreen():
            while True:
                try:
                    query_start = time.time()
                    table = self.format(args, tqdm_disable=True).split('\n')
                    table = f'{term.clear_eol}\n'.join(table)
                    print(term.move(0, 0), end='')
                    dt = datetime.fromtimestamp(query_start)
                    spaces = ' ' * 12
                    header = f'{platform.node()}{spaces}{dt:%Y-%m-%d %H:%M:%S}'
                    print(header + term.clear_eol)
                    if table:
                        print(table, end=term.clear_eol)
                    print(term.clear_eos, end='')
                    sys.stdout.flush()
                    query_duration = time.time() - query_start
                    sleep_duration = args.interval - query_duration
                    if sleep_duration > 0:
                        time.sleep(sleep_duration)
                except KeyboardInterrupt:
                    return 0

    def main(self, args):
        if args.interval is not None:
            return self.loop(args)
        print(self.format(args, tqdm_disable=False))
        return 0


@register_action(
    'list', 'show job infos in compact format',
    aliases=['ls'], default=True)
class ListAction(ReadActionBase):
    list_options = {
        ('-l', '--length'): {
            'type': int,
            'default': 30,
            'help': 'Max length for each column.',
        },
        ('-c', '--columns'): {
            'type': str,
            'default': 'id,status,start,time,command',
            'help':
                'Columns to display, separated by commas. Available columns: '
                'id, status, slots, gpus, gpu_ids, '
                'enqueue, start, end, time, command, output. '
                'Alternatively, prefix a column with "+" to show it, '
                'and with "-" to hide it.'
        },
        ('-t', '--table-format'): {
            'type': str,
            'default': 'rounded_outline',
            'help': 'The format of the table.',
        },
    }

    def __init__(self, name, parser_kwargs):
        super().__init__(name, parser_kwargs)
        self.options.update(self.list_options)

    @staticmethod
    def shorten(text, max_len):
        if not max_len:
            return text
        return text if len(text) <= max_len else text[:max_len - 3] + '...'

    @staticmethod
    def _color_status(status):
        return COLOR_STATUS[status](status).replace('\x1b(B', '')

    @staticmethod
    def _format_time(dt):
        return dt.strftime('%m-%d %H:%M') if dt else None

    def format(self, args, tqdm_disable=False):
        info = self.backend.full_info(
            self.ids, self.filters, tqdm_disable=tqdm_disable)
        if not info:
            return 'No jobs found.'
        rows = []
        default_columns = ['id', 'status', 'start', 'time', 'command']
        columns = [c.lower() for c in args.columns.split(',')]
        plus_columns = [c for c in columns if c.startswith('+')]
        minus_columns = [c for c in columns if c.startswith('-')]
        columns = [c for c in columns if c not in plus_columns + minus_columns]
        if columns and (plus_columns or minus_columns):
            print('Cannot mix +/- columns with explicit columns.')
            return 1
        else:
            columns = default_columns + [c[1:] for c in plus_columns]
            columns = [c for c in columns if f'-{c}' not in minus_columns]
        columns = list(dict.fromkeys(columns))
        for i in info:
            try:
                time_run = timedelta_format(i['time_run'], 'wdhms', 2)
            except KeyError:
                time_run = ''
            row = {
                'ID': i['id'],
                'Status': self._color_status(i['status']),
                'Slots': i.get('slots_required', 1),
                'GPUs': i.get('gpus_required', 0),
                'GPU IDs': i.get('gpu_ids', ''),
                'Enqueue': self._format_time(i.get('enqueue_time')),
                'Start': self._format_time(i.get('start_time')),
                'End': self._format_time(i.get('end_time')),
                'Time': time_run,
                'Command': self.shorten(i['command'], args.length),
            }
            if 'output' in columns:
                if i['status'] not in ['queued']:
                    out = self.backend.output(i, 1)
                else:
                    out = ''
                row['Output'] = self.shorten(out, args.length)
            rows.append({
                k: v for k, v in row.items()
                if k.replace(' ', '_').lower() in columns})
        maxcolwidths = [
            None if c not in ['output', 'command'] else args.length
            for c in columns]
        table = tabulate.tabulate(
            rows, headers='keys', tablefmt=args.table_format,
            maxcolwidths=maxcolwidths)
        return table


@register_action('ids', help='show job IDs', aliases=['id'])
class IdsAction(ReadActionBase):
    def format(self, args, tqdm_disable=False):
        jobs = self.backend.job_info(self.ids, self.filters)
        return ', '.join(str(i['id']) for i in jobs)


@register_action('info', 'show job infos')
class InfoAction(ReadActionBase):
    def format(self, args, tqdm_disable=False):
        info = self.backend.full_info(
            self.ids, self.filters, tqdm_disable=tqdm_disable)
        if not info:
            return 'No jobs found.'
        outputs = []
        for i in info:
            outputs.append(f'Job {i["id"]}:')
            for k, v in i.items():
                if k == 'id':
                    continue
                outputs.append(f'  {k}: {v}')
        return '\n'.join(outputs)


@register_action('commands', 'show job commands', aliases=['cmd'])
class CommandsAction(ReadActionBase):
    commands_options = {
        ('-j', '--no-job-ids'): {
            'action': 'store_true',
            'help': 'Do not print the job ID before the command.',
        },
    }

    def __init__(self, name, parser_kwargs):
        super().__init__(name, parser_kwargs)
        self.options.update(self.commands_options)

    def format(self, args, tqdm_disable=False):
        info = self.backend.full_info(
            self.ids, self.filters, tqdm_disable=tqdm_disable)
        outputs = []
        for i in info:
            if args.no_job_ids:
                outputs.append(i['command'])
            else:
                outputs.append(f"{i['id']}: {i['command']}")
        return '\n'.join(outputs)


@register_action(
    'export', 'show job infos in machine-readable format', aliases=['e'])
class ExportAction(ReadActionBase):
    export_options = {
        ('-e', '--export-format'): {
            'choices': ['json', 'yaml', 'toml'],
            'default': 'json',
            'help': 'The format of the output.',
        },
        ('-t', '--tail'): {
            'type': int,
            'default': None,
            'help':
                'The tail lines of the output. '
                'If 0, all lines will be shown.',
        },
    }

    def __init__(self, name, parser_kwargs):
        super().__init__(name, parser_kwargs)
        self.options.update(self.export_options)

    def extra_func(self, i, args):
        i['time_run'] = i['time_run'].total_seconds()
        if args.tail is not None:
            i['output'] = self.backend.output(i, args.tail)

    def format(self, args, tqdm_disable=False):
        extra_func = functools.partial(self.extra_func, args=args)
        info = self.backend.full_info(
            self.ids, self.filters,
            extra_func=extra_func, tqdm_disable=tqdm_disable)
        if args.export_format == 'json':
            import json
            return json.dumps(info, indent=4, default=str)
        elif args.export_format == 'yaml':
            import yaml
            return yaml.dump(info, default_flow_style=False)
        elif args.export_format == 'toml':
            import toml
            return toml.dumps({'job': info})
        else:
            raise ValueError(f'Unknown format: {args.export_format}')


@register_action('outputs', 'show job outputs', aliases=['out'])
class OutputsAction(ReadActionBase):
    outputs_options = {
        ('-i', '--interactive'): {
            'action': 'store_true',
            'help': 'Follow the output.',
        },
        ('-t', '--tail'): {
            'type': int,
            'default': 0,
            'help':
                'Number of lines. '
                'If 0, all lines will be shown.',
        },
        ('-R', '--raw'): {
            'action': 'store_true',
            'help': 'Do not format lines.',
        },
    }

    def __init__(self, name, parser_kwargs):
        super().__init__(name, parser_kwargs)
        self.options.update(self.outputs_options)

    def format(self, args, tqdm_disable=False):
        info = self.backend.job_info(self.ids, self.filters)
        if not info:
            return 'No jobs found.'
        outputs = []
        for i in info:
            if i['status'] in ['queued']:
                continue
            outputs.append(f'Job {i["id"]}:')
            out = self.backend.output(i, args.tail, shell=False)
            if not args.raw:
                out = textwrap.indent(out, '> ') + '\n'
            outputs.append(out)
        return '\n'.join(outputs).rstrip()

    def main(self, args):
        if not args.interactive:
            return super().main(args)
        info = self.backend.job_info(self.ids, self.filters)
        if not info:
            print('No jobs found.')
            return 1
        if len(info) > 1:
            print('Cannot follow multiple outputs.')
            return 1
        try:
            self.backend.output(info[0], args.tail, shell=True)
        except KeyboardInterrupt:
            pass
        return 0


@register_action('wait', 'wait for jobs to finish', aliases=['w'])
class WaitAction(ReadActionBase):
    wait_options = {
        ('-p', '--progress'): {
            'action': 'store_true',
            'help': 'Show progress bar.',
        },
    }

    def __init__(self, name, parser_kwargs):
        super().__init__(name, parser_kwargs)
        self.options.update(self.wait_options)

    def main(self, args, tqdm_disable=False):
        f = FilterArgs(running=True, queued=True)
        info = self.backend.job_info(self.ids, f)
        pbar = tqdm(total=len(info), desc='wait') if args.progress else None
        while True:
            remaining = self.backend.job_info(self.ids, f)
            if pbar:
                pbar.n = len(info) - len(remaining)
                pbar.refresh()
            time.sleep(1)
            if not remaining:
                break
        if pbar:
            pbar.close()
            print('')
        return 0



import re
import shlex
import datetime

from ..common import tqdm, STATUSES, file_tail_lines, tail_lines, dict_simplify
from .base import register_backend, BackendBase


@register_backend('ts')
class TaskSpoolerBackend(BackendBase):
    def __init__(self, name, config):
        super().__init__(name, config)
        socket = self.config.get('socket', 'default')
        slots = self.config.get('slots')
        self.env.setdefault('TS_SLOTS', str(slots))
        self.env.setdefault('TS_SOCKET', f'/tmp/ts-{socket}.sock')
        self.env = dict_simplify(self.env, not_value=True)
        self.backend_getset('slots', slots)

    def backend_getset(self, key, value=None):
        slot_keys = [
            'slots',
            f'backends.{self.name}.slots',
            f'groups.{self.config["group"]}.slots'
        ]
        if key in slot_keys:
            if value is None:
                return int(self.exec('-S') or -1)
            return self.exec('-S', value)

    def backend_info(self):
        info = {
            'name': 'Task Spooler',
            'command': self._command,
        }
        version = self.exec('-V') or ''
        result = re.search(r'(v[\.\d]+)', version)
        if result:
            info['version'] = result.group(1)
        return info

    def backend_kill(self, args):
        self.exec('-K')

    def job_info(self, ids=None, filters=None):
        info = []
        tsout = self.exec()
        if not tsout:
            return info
        for l in tsout.splitlines()[1:]:
            l = l.strip().split()
            if l[1] == 'finished':
                job_id, status, _, exitcode, *_ = l
                exitcode = int(exitcode)
                if exitcode == 0:
                    status = 'success'
                elif exitcode < 0:
                    status = 'killed'
                else:
                    status = 'failed'
            else:
                job_id, status, *_ = l
                exitcode = None
            if status == 'allocating':
                status = 'queued'
            info.append({
                'id': int(job_id),
                'status': status,
                'exitcode': exitcode
            })
        if filters is not None and not filters.all:
            for a in STATUSES:
                if getattr(filters, a):
                    continue
                info = [i for i in info if i['status'] != a]
        if ids is not None:
            info = [i for i in info if i['id'] in ids]
        return info

    @classmethod
    def get_line(cls, ji, key):
        for l in ji.splitlines():
            if key in l:
                return l.replace(key, '').strip()
        return None

    @classmethod
    def get_time(cls, ji, key):
        time = cls.get_line(ji, key)
        if not time:
            return None
        return datetime.datetime.strptime(time, '%a %b %d %H:%M:%S %Y')

    def full_info(
        self, ids=None, filters=None, extra_func=None, tqdm_disable=False
    ):
        info = self.job_info(ids, filters)
        for i in tqdm(info, disable=tqdm_disable, desc='info'):
            ji = self.exec('-i', i['id'])
            start_time = self.get_time(ji, 'Start time: ')
            end_time = self.get_time(ji, 'End time: ')
            if not start_time:
                delta = None
            elif not end_time:
                delta = datetime.datetime.now() - start_time
            else:
                delta = end_time - start_time
            try:
                pid = int(self.exec('-p', i['id'], check=False) or -1)
            except ValueError:
                pid = None
            new_info = {
                'command': self.get_line(ji, 'Command: '),
                'slots_required':
                    int(self.get_line(ji, 'Slots required: ') or 1),
                'gpus_required':
                    int(self.get_line(ji, 'GPUs required: ') or 0),
                'gpu_ids': self.get_line(ji, 'GPU IDs: '),
                'enqueue_time': self.get_time(ji, 'Enqueue time: '),
                'start_time': start_time,
                'end_time': end_time,
                'time_run': delta,
                'output_file': self.exec('-o', i['id'], check=False),
                'pid': pid,
            }
            i.update({k: v for k, v in new_info.items() if v is not None})
            if extra_func:
                extra_func(i)
        return info

    def output(self, info, tail, shell=False):
        if shell:
            self.exec('-c', info['id'], check=False, shell=True)
            return ''
        if info['status'] in ['success', 'failed', 'killed']:
            return tail_lines(self.exec('-c', info['id'], check=False), tail)
        f = info.get('output_file') or self.exec('-o', info['id'], check=False)
        return file_tail_lines(f, tail) if f else ''

    def add(self, command, gpus=None, slots=None, commit=True):
        torun = []
        alloc_config = self.config.get('alloc', {})
        gpus = gpus if gpus is not None else alloc_config.get('gpus', 1)
        slots = slots if slots is not None else alloc_config.get('slots', 1)
        if gpus > 0:
            torun += ['-G', gpus]
        torun += ['-N', slots]
        command = command.replace('\n', '\\n')
        torun += shlex.split(command)
        return self.exec(*torun, commit=commit)

    def kill(self, info, commit=True):
        self.exec('-k', info['id'], commit=commit)

    def remove(self, info, commit=True):
        if info['status'] == 'running':
            self.kill(info, commit=commit)
            self.exec('-w', info['id'], commit=commit, check=False)
        return self.exec('-r', info['id'], commit=commit)



import os
import sys
import argparse

import tomlkit

from . import __version__
from .common import dict_merge
from .actions import INFO
from .backends import BACKENDS


class CLI:
    base_options = {
        ('-V', '--version'): {
            'action': 'store_true',
            'help': 'print the version and exit',
        },
        ('-rc', '--rc-file'): {
            'type': str,
            'default': None,
            'help':
                'The configuration file to use. '
                'If not provided, it reads from "~/.config/tq.toml" '
                'and "./.tq.toml".'
        },
        ('-b', '--backend'): {
            'type': str,
            'default': None,
            'choices': list(BACKENDS.keys()),
            'help': 'The backend to use.',
        },
        ('-g', '--group'): {
            'type': str,
            'default': None,
            'help': 'The group to use.',
        },
    }
    rc_path = '.tq.toml'

    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser()
        for option, kwargs in self.base_options.items():
            self.parser.add_argument(*option, **kwargs)
        action_parsers = self.parser.add_subparsers(dest='action')
        for action in INFO['actions'].values():
            kwargs = action.parser_kwargs
            action_parser = action_parsers.add_parser(**kwargs)
            for option, kwargs in action.options.items():
                action_parser.add_argument(*option, **kwargs)

    def _load_config(self, args):
        if args.rc_file:
            with open(args.rc_file, 'r', encoding='utf-8') as f:
                try:
                    return tomlkit.load(f)
                except tomlkit.exceptions.ParseError as e:
                    print(f'Error parsing {args.rc_file}: {e}')
                    return {}
        else:
            args.rc_file = self.rc_path
        rc_files = [
            os.path.join(os.path.dirname(__file__), '..', 'default.toml'),
            '~/.config/tq.toml',
            self.rc_path,
        ]
        config = {}
        for path in rc_files:
            try:
                with open(path, 'r', encoding='utf-8') as f:
                    try:
                        rc = tomlkit.load(f)
                    except tomlkit.exceptions.ParseError as e:
                        print(f'Error parsing {path}: {e}')
                        continue
                config = dict_merge(config, rc)
            except FileNotFoundError:
                pass
        return config

    def _resolve_config(self, args, config):
        backend = args.backend or config.get('backend', 'ts')
        group = args.group or config.get('group', 'default')
        backend_config = config.get('backends', {}).get(backend, {})
        group_config = config.get('groups', {}).get(group, {})
        dict_merge(config, backend_config)
        dict_merge(config, group_config)
        config.update({
            'backend': backend,
            'group': group,
        })
        return config

    def main(self, args=None):
        args = args or sys.argv[1:]
        if all(a not in INFO['aliases'] for a in args):
            if '-h' in args or '--help' in args:
                self.parser.print_help()
                sys.exit(0)
            if '-V' in args or '--version' in args:
                print(__version__)
                sys.exit(0)
            args = [INFO['default']] + args
        args = self.parser.parse_args(args)
        config = self._load_config(args)
        config = self._resolve_config(args, config)
        try:
            action = INFO['aliases'][args.action]
            action_func = INFO['actions'][action]
        except KeyError:
            print(f'Invalid action: {args.action}')
            sys.exit(1)
        return action_func(args, config)


def main():
    sys.exit(CLI().main())


if __name__ == '__main__':
    main()



import copy
import tomlkit

from ..common import dict_simplify
from .base import register_action, ActionBase


@register_action('config', 'config action', aliases=['c'])
class ConfigAction(ActionBase):
    config_options = {
        ('key', ): {
            'type': str,
            'nargs': '?',
            'help': 'The key to get/set.',
        },
        ('value', ): {
            'type': str,
            'nargs': '?',
            'help': 'The value to set.  Use "null" to delete.',
        },
    }

    def __init__(self, name, parser_kwargs):
        super().__init__(name, parser_kwargs)
        self.options.update(self.config_options)

    def getset(self, mapping, key, value=None):
        *key_parts, last_key = key.split('.')
        for k in key_parts:
            mapping = mapping.setdefault(k, {})
        if value is None:
            bvalue = self.backend.backend_getset(key)
            if bvalue is not None:
                return bvalue
            return mapping.get(last_key)
        if value == 'null':
            del mapping[last_key]
        else:
            mapping[last_key] = value
        self.backend.backend_getset(key, value)

    def main(self, args):
        config = dict_simplify(copy.deepcopy(self.backend.config))
        if args.key is None:
            print(tomlkit.dumps(config).rstrip())
            return
        if args.value is None:
            print(self.getset(config, args.key))
            return
        try:
            with open(args.rc_file, 'r', encoding='utf-8') as f:
                rc_config = tomlkit.load(f)
        except FileNotFoundError:
            rc_config = {}
        self.getset(rc_config, args.key, args.value)
        with open(args.rc_file, 'w', encoding='utf-8') as f:
            f.write(tomlkit.dumps(rc_config))



import shlex
import base64

from ..base import register_backend, BackendBase


@register_backend('screen')
class ScreenBackend(BackendBase):
    def __init__(self, name, config):
        super().__init__(name, config)
        socket = self.config.get('socket', 'default')
        self.env.setdefault('SCREENDIR', f'/tmp/screen-{socket}.sock')

    def backend_getset(self, key, value=None):
        pass

    def backend_info(self):
        raise NotImplementedError

    def backend_kill(self, args):
        raise NotImplementedError

    def backend_command(self, command, commit=True):
        raise NotImplementedError

    def job_info(self, ids=None, filters=None):
        raise NotImplementedError

    def full_info(
        self, ids=None, filters=None, extra_func=None, tqdm_disable=False
    ):
        print(self.exec('-ls', check=False))

    def output(self, info, tail, shell=False):
        raise NotImplementedError

    def add(self, command, gpus, slots, commit=True):
        key = base64.b64encode(command.encode()).decode().rstrip('=')
        torun = ['-dmS', key] + shlex.split(command)
        print(self.exec(*torun, commit=commit))

    def kill(self, info, commit=True):
        raise NotImplementedError

    def remove(self, info, commit=True):
        raise NotImplementedError



from .base import register_backend, BackendBase


@register_backend('dummy')
class DummyBackend(BackendBase):
    def __init__(self, name, config):
        config['command'] = None
        super().__init__(name, config)

    def backend_info(self):
        return {
            'name': 'dummy',
            'command': '',
            'version': '0.0.0',
        }

    def backend_kill(self, args):
        print(args)

    def backend_command(self, command, commit=True):
        return command, commit

    def job_info(self, ids=None, filters=None):
        print(ids, filters)
        return []

    def full_info(
        self, ids=None, filters=None, extra_func=None, tqdm_disable=False
    ):
        print(ids, filters)
        return []

    def output(self, info, tail, shell=False):
        print(info, tail, shell)

    def add(self, command, gpus, slots, commit=True):
        print(repr(command), gpus, slots, commit)
        return 0

    def kill(self, info, commit=True):
        print(info, commit)

    def remove(self, info, commit=True):
        print(info, commit)



import os
import argparse

import flask


class ScreenBackendServer:
    def __init__(self, screen_dir):
        super().__init__()
        self.screen_dir = screen_dir

    def poll(self):
        ...


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('action', choices=['serve', ''])


if __name__ == '__main__':



import sys

from ..common import tqdm
from .base import register_action, DryActionBase
from .filter import FilterActionBase



class WriteActionBase(FilterActionBase, DryActionBase):
    def transform_args(self, args):
        args = DryActionBase.transform_args(self, args)
        args = FilterActionBase.transform_args(self, args)
        if not self.has_filters:
            print(
                'tsu: flags or ids must be specified '
                f'for dangerous action {self.name!r}.')
            sys.exit(1)
        return args


@register_action('kill', 'kill jobs', aliases=['k'])
class KillAction(WriteActionBase):
    def kill(self, info, commit):
        for i in tqdm(info, desc='kill'):
            self.backend.kill(i, commit=commit)

    def main(self, args):
        info = self.backend.job_info(self.ids, self.filters)
        info = [i for i in info if i['status'] == 'running']
        if not info:
            print('No job to kill.')
            return
        self.kill(info, args.commit)
        killed_ids = ', '.join(str(i['id']) for i in info)
        print('Killed:', killed_ids)


@register_action('remove', 'remove jobs', aliases=['rm'])
class RemoveAction(WriteActionBase):
    def remove(self, info, commit):
        queued = [i for i in info if i['status'] == 'queued']
        remaining = [i for i in info if i['status'] != 'queued']
        for i in tqdm(queued + remaining, desc='remove'):
            self.backend.remove(i, commit=commit)

    def main(self, args):
        info = self.backend.job_info(self.ids, self.filters)
        if not info:
            print('No job to remove.')
            return
        self.remove(info, args.commit)
        removed_ids = ', '.join(str(i['id']) for i in info)
        print('Removed:', removed_ids)


@register_action('rerun', 'rerun jobs', aliases=['rr'])
class RerunAction(WriteActionBase):
    def rerun(self, info, commit):
        new_ids = []
        for i in tqdm(info, desc='rerun'):
            command = i['command']
            gpus = i['gpus_required']
            slots = i['slots_required']
            ji = self.backend.add(command, gpus, slots, commit=commit)
            new_ids.append(ji)
        return new_ids

    def main(self, args):
        info = self.backend.full_info(self.ids, self.filters)
        if not info:
            print('No job to rerun.')
            return
        new_ids = self.rerun(info, args.commit)
        reran_ids = ', '.join(
            f"{i['id']} -> {j}" for i, j in zip(info, new_ids))
        print('Reran:', reran_ids)


@register_action('requeue', 'requeue jobs', aliases=['rq'])
class RequeueAction(RerunAction, RemoveAction):
    def main(self, args):
        info = self.backend.full_info(self.ids, self.filters)
        if not info:
            print('No job to requeue.')
            return
        new_ids = self.rerun(info, args.commit)
        self.remove(info, args.commit)
        requeued_ids = ', '.join(
            f"{i['id']} -> {j}" for i, j in zip(info, new_ids))
        print('Requeued:', requeued_ids)



import argparse

import tomlkit

from ..common import tqdm, FilterArgs
from .base import register_action, ActionBase


@register_action('backend', 'backend actions', aliases=['b'])
class BackendAction(ActionBase):
    backend_options = {
        ('backend_action', ): {
            'type': str,
            'default': None,
            'choices': ['info', 'reset', 'command'],
            'help': 'The action to perform.',
        },
        ('command', ): {
            'type': str,
            'nargs': argparse.REMAINDER,
            'help': 'The command to run.',
        },
    }

    def __init__(self, name, parser_kwargs):
        super().__init__(name, parser_kwargs)
        self.options.update(self.backend_options)

    def info(self, args):
        binfo = self.backend.backend_info()
        print(tomlkit.dumps(binfo).rstrip())

    def reset(self, args):
        info = self.backend.full_info(None, FilterArgs())
        for i in tqdm(info, desc='remove'):
            self.backend.remove(i)
        self.backend.backend_kill(args)
        print(f'Killed {self.backend.name} backend.')

    def command(self, args):
        if not args.command:
            print('No command provided.')
            return 1
        print(self.backend.backend_command(args.command))

    def main(self, args):
        try:
            func = getattr(self, args.backend_action)
        except AttributeError:
            print(f'Invalid backend action: {args.backend_action}')
            return 1
        return func(args)



import sys

from ..common import STATUSES, FilterArgs
from .base import ActionBase


class FilterActionBase(ActionBase):
    filter_options = {
        ('id', ): {
            'type': str,
            'default': None,
            'nargs': '?',
            'help':
                'Optional ranges of job IDs to perform the action on, '
                'a comma-separated list of ranges, e.g. "1-3,5,7-9". '
                'If not provided, all jobs will be affected. '
                'If only "-" is specified, it reads from stdin.'
        },
        ('-A', '--all'): {
            'action': 'store_true',
            'help': 'Perform the action on all jobs.',
        },
        ('-r', '--running'): {
            'action': 'store_true',
            'help': 'Perform the action on running jobs.',
        },
        ('-q', '--queued'): {
            'action': 'store_true',
            'help': 'Perform the action on queued/allocating jobs.',
        },
        ('-s', '--success'): {
            'action': 'store_true',
            'help': 'Perform the action on successful jobs.',
        },
        ('-f', '--failed'): {
            'action': 'store_true',
            'help': 'Perform the action on failed jobs.',
        },
        ('-k', '--killed'): {
            'action': 'store_true',
            'help': 'Perform the action on killed jobs.',
        },
    }

    def __init__(self, name, parser_kwargs):
        super().__init__(name, parser_kwargs)
        self.options.update(self.filter_options)

    def _parse_ids(self, args):
        if args.id == '-':
            args.id = sys.stdin.read().strip()
            return
        if not args.id:
            self.ids = None
            return
        ids = []
        for i in args.id.split(','):
            if '-' in i:
                start, end = i.split('-')
                ids += list(range(int(start), int(end) + 1))
            else:
                ids.append(int(i))
        self.ids = ids

    def _parse_filters(self, args):
        self.filters = FilterArgs(
            force_all=args.all,
            running=args.running,
            queued=args.queued,
            success=args.success,
            failed=args.failed,
            killed=args.killed,
        )

    def transform_args(self, args):
        self._parse_ids(args)
        self._parse_filters(args)
        return args

    @property
    def has_filters(self):
        filters = [self.ids, self.filters.force_all]
        filters += [getattr(self.filters, a) for a in STATUSES]
        return any(filters)
